//***************************************************************************
// Group Common Stuff
// File common.cc
// Date 04.11.06 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//***************************************************************************

#include <sys/time.h>
#include <stdarg.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#include <common.hpp>

int theEloquence;

//***************************************************************************
// Tell
//***************************************************************************

int tell(int eloquence, const char* format, ...)
{
   const int sizeTime = 8;        // "12:12:34"
   const int sizeMSec = 4;        // ",142"
   const int sizeHeader = sizeTime + sizeMSec + 1;
   const int maxBuf = 1000;

   struct timeval tp;
   char buf[maxBuf+TB];
   va_list ap;
   time_t now;

   va_start(ap, format);

   if (theEloquence >= eloquence)
   {
      time(&now);
      gettimeofday(&tp, 0);

      vsnprintf(buf + sizeHeader, maxBuf - sizeHeader, format, ap);
      strftime(buf, sizeTime+1, "%H:%M:%S", localtime(&now));

      sprintf(buf+sizeTime, ",%3.3ld", tp.tv_usec / 1000);
      
      buf[sizeHeader-1] = ' ';
      printf("%s\n", buf);
   }

   va_end(ap);
   
   return 0;
}

//***************************************************************************
// Not Null
//***************************************************************************

const char* notNull(const char* s)
{ 
   return s ? s : ""; 
}

//***************************************************************************
// Class SlotService
//***************************************************************************
//***************************************************************************
// Is Bit
//***************************************************************************

int SlotService::isBit(unsigned int value, int bit)
{
//    tell(eloDebug, "Debug: isBit(%u, %d) mask is %d, result is %d", 
//         value, bit, (1<<bit), value & (1 << bit));

   return value & (1 << bit);
}

//***************************************************************************
// Toggle Bit
//***************************************************************************

void SlotService::toggleBit(unsigned int& value, int bit)
{
   setBitTo(value, bit, !isBit(value, bit));
}

//***************************************************************************
// Set Bit To
//***************************************************************************

void SlotService::setBitTo(unsigned int& value, int bit, int state)
{
   if (state)
      value |= 1 << bit;
   else
      value &= ~(1 << bit);
}

//***************************************************************************
// To Bin Str
//***************************************************************************

const char* SlotService::toBinStr(unsigned int value, char* mask, int bits)
{
   // int bits = bytes * 8;
   int bytes = bits / 8;
   int bit = 1;

   if (bytes * 8 < bits)
      bytes++;

   for (int i = 0, b = 0; b < bits; i++, b++)
   {
      if (!(b % 8) && b)
      {
         mask[bits+bytes-i-2] = ':';
         i++;
      }
      
      mask[bits+bytes-i-2] = (value & bit) ? '1' : '0';     
      bit = bit << 1;
   }

   mask[bits+bytes-1] = 0;

   return mask;
}

//***************************************************************************
// kmh
//***************************************************************************

double SlotService::kmh(double meter, double usec)
{
   return ((double)meter/1000L)  /  
      ((double)usec / 1000000L / 60L / 60L);
}

//***************************************************************************
// Init TV
//***************************************************************************

void SlotService::tvNull(timeval* tv)
{
   memset(tv, 0, sizeof(timeval));
}

timeval* SlotService::tvNow(timeval* tv)
{
   gettimeofday(tv, 0);
   return tv;
}

//***************************************************************************
// 
//***************************************************************************

timeval SlotService::addMs2Tv(timeval tv, unsigned long milli)
{
   timeval tvSum;
   unsigned long usec = tv.tv_usec + (milli % 1000) * 1000;

   tvSum.tv_usec = usec % 1000000;
   tvSum.tv_sec  = tv.tv_sec + (milli / 1000) + (usec / 1000000);

   return tvSum;
}

//***************************************************************************
// Elapsed in Seconds
//***************************************************************************

long SlotService::elapsed(const timeval* tpFrom, const timeval* tpNow)
{
   int sec = tpNow->tv_sec - tpFrom->tv_sec;
   int usec = sec*1000000 + tpNow->tv_usec;

   usec -= tpFrom->tv_usec;

   return usec;
}

//***************************************************************************
// Seconds String
//***************************************************************************

const char* SlotService::secStr(long usec, char* buf)
{
   sprintf(buf, "%2ld.%06ld", usec/1000000, usec%1000000);

   return buf;
}

//***************************************************************************
// To Str
//***************************************************************************

const char* SlotService::toStr(int value)
{
   static char buf[10+TB];

   sprintf(buf, "%d", value);

   return buf;
}
