//***************************************************************************
// Group Linslot / Linux - Slotrace Manager
// File roundprofile.cc
// Date 03.03.08 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//***************************************************************************

//***************************************************************************
// Include
//***************************************************************************

#include <common.hpp>
#include <roundprofile.hpp>

//***************************************************************************
// RenderArea
//***************************************************************************

RenderArea::RenderArea(QWidget* parent)
   : QWidget(parent)
{
   setBackgroundRole(QPalette::Base);
   setAutoFillBackground(true);

   xScale = 10;
}

//***************************************************************************
// 
//***************************************************************************

void RenderArea::scaleUpClicked(bool)
{
   xScale++;
   repaint();
}

void RenderArea::scaleDownClicked(bool)
{
   if (xScale > 5)
      xScale--;

   repaint();
}

//***************************************************************************
// Paint Event
//***************************************************************************

void RenderArea::paintEvent(QPaintEvent* )
{
   int x, y;

   QPainter painter(this);
   painter.setPen(pen);
   painter.setBrush(brush);

   int xNull = 30;
   int xWidth = width() - 2*xNull;

   int yNull = height() - 30;
   int yHeight = height() - 2*30;
   int yScale = (int)(((double)yHeight/100.0) * 2.0);   // alle 5% ein step

   // draw x axis

   painter.drawLine(xNull, yNull, xNull + xWidth, yNull);

   for (x = 0; x < xWidth; x += xScale)
   {
      if (!(x % (10*xScale)))
         painter.drawLine(xNull + x, yNull-4, xNull + x, yNull+4);
      else
         painter.drawLine(xNull + x, yNull-2, xNull + x, yNull+2);
   }

   // draw y axis

   painter.drawLine(xNull, yNull, xNull, yNull-yHeight);

   for (y = 0; y < yHeight-yScale; y += yScale)
   {
      if (!(y % (5*yScale)))
         painter.drawLine(xNull-4, yNull-y, xNull+4, yNull-y);
      else
         painter.drawLine(xNull-2, yNull-y, xNull+2, yNull-y);
   }

   // draw frame rect

   painter.setPen(palette().dark().color());
   painter.setBrush(Qt::NoBrush);
   painter.drawRect(QRect(0, 0, width() - 1, height() - 1));
}
